"""The Chain runs a self-review of logical fallacies as determined by this paper  \
categorizing and defining logical fallacies https://arxiv.org/pdf/2212.07425.pdf. \
Modeled after Constitutional AI and in same format, but applying logical \
fallacies as generalized rules to remove in output"""
