"""HuggingFace Security toolkit."""

from langchain_experimental.prompt_injection_identifier.hugging_face_identifier import (
    HuggingFaceInjectionIdentifier,
)

__all__ = ["HuggingFaceInjectionIdentifier"]
